from copy import copy
from datetime import timedelta

import pydash as _
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import Client, RequestFactory, TestCase, override_settings
from django.utils import timezone

from acserv import proto
from acserv.helpers import base64_to_proto
from acserv.views import ACView, IndexACView

User = get_user_model()

INDEX_BODY = (b'<?xml version="1.0" encoding="UTF-8"?>\n'
              b'<config-auth client="vpn" type="init" aggregate-auth-version="2">\n'
              b'<version who="vpn">4.0.05069</version>\n'
              b'<device-id device-type="iPhone8,1" platform-version="10.3.3" unique-id="0b1726e825960936d76adf6caf004b97e35a7760">apple-ios</device-id>\n'
              b'<phone-id>unknown</phone-id>\n'
              b'<group-select>MacBookPro11,3(mac-intel 10.12.1)</group-select>\n'
              b'<group-access>https://192.168.31.244</group-access>\n'
              b'</config-auth>\n')

INDEX_META = {
    'CONTENT_TYPE': 'application/x-www-form-urlencoded',
    'HTTP_ACCEPT': '*/*',
    'HTTP_ACCEPT_ENCODING': 'identity',
    'HTTP_CONNECTION': 'close,close',
    'HTTP_HOST': '192.168.31.244',
    'HTTP_PROXY': '192.168.31.244',
    'HTTP_USER_AGENT': 'AnyConnect AppleSSLVPN_Darwin_ARM (iPhone) 4.0.05069',
    'HTTP_X_AGGREGATE_AUTH': '1',
    'HTTP_X_ANYCONNECT_IDENTIFIER_CLIENTVERSION': '4.0.05069',
    'HTTP_X_ANYCONNECT_IDENTIFIER_DEVICETYPE': 'iPhone8,1',
    'HTTP_X_ANYCONNECT_IDENTIFIER_DEVICE_IMEI': 'unknown',
    'HTTP_X_ANYCONNECT_IDENTIFIER_DEVICE_MACADDRESS': 'unknown',
    'HTTP_X_ANYCONNECT_IDENTIFIER_DEVICE_UNIQUEID': '0b1726e825960936d76adf6caf004b97e35a7760',
    'HTTP_X_ANYCONNECT_IDENTIFIER_PLATFORM': 'apple-ios',
    'HTTP_X_ANYCONNECT_IDENTIFIER_PLATFORMVERSION': '10.3.3',
    'HTTP_X_FORWARDED_FOR': '10.0.2.2',
    'HTTP_X_PASSCODE': 'hello',
    'HTTP_X_TRANSCEND_VERSION': '1,1',
    'QUERY_STRING': '',
    'REMOTE_ADDR': '192.168.31.244',
    'REMOTE_PORT': '55192',
}

USERNAME_BODY = b'username=a%40a.aa'
USERNAME_META = copy(INDEX_META)
USERNAME_META.update({
    'HTTP_COOKIE': 'webvpnc=bu:/&p:t&iu:1/&sh:369B0AE7D129B6F0B74616340D96EA2D74FA9747; '
                   'webvpncontext=CiQzZjQyODYyNi1kMTNhLTQ1NGMtYTk2NS0zNjA4ODA5MmY4NGEaKDBiMTcyNmU4MjU5NjA5MzZkNzZhZGY2Y2FmMDA0Yjk3ZTM1YTc3NjAiCWlQaG9uZTgsMTIHdW5rbm93bjoJYXBwbGUtaW9zQgYxMC4zLjNKCTQuMC4wNTA2OVI0QW55Q29ubmVjdCBBcHBsZVNTTFZQTl9EYXJ3aW5fQVJNIChpUGhvbmUpIDQuMC4wNTA2OQ==',
})

PASSWORD_BODY = b'password=password'


def override_auth_header(val):
    return _.merge({}, settings.ANYCONNECT, {'AUTH_HEADER': val})


@override_settings(ANYCONNECT=override_auth_header({'auth-header': 'hello'}))
class ACViewTest(TestCase):
    def test_make_auth_context(self):
        req = RequestFactory().post('/',
                                    data=INDEX_BODY,
                                    content_type="application/x-www-form-urlencoded",
                                    HTTP_AUTH_HEADER='hello',
                                    **INDEX_META)
        req.session = {}
        ctx = IndexACView.make_auth_context(req)

        self.assertIsNotNone(ctx.uid, '随机生成')
        self.assertEqual(ctx.device_id, "0b1726e825960936d76adf6caf004b97e35a7760")
        self.assertEqual(ctx.device_type, "iPhone8,1")
        self.assertEqual(ctx.device_imei, "unknown")
        self.assertEqual(ctx.platform, "apple-ios")
        self.assertEqual(ctx.platform_version, "10.3.3")
        self.assertEqual(ctx.client_version, "4.0.05069")
        self.assertEqual(ctx.user_agent, "AnyConnect AppleSSLVPN_Darwin_ARM (iPhone) 4.0.05069")

    def test_auth_header(self):
        req = RequestFactory().get('/')
        resp = ACView.as_view()(req)
        self.assertEqual(resp.status_code, 403)

        req = RequestFactory().get('/', HTTP_AUTH_HEADER='hello')
        resp = ACView.as_view()(req)
        self.assertEqual(resp.status_code, 401, '带合法头就不会403')

    def test_get_auth_context_or_response(self):
        req = RequestFactory().post('/',
                                    data=INDEX_BODY,
                                    content_type="application/x-www-form-urlencoded",
                                    **INDEX_META)
        auth_ctx, resp = ACView.get_auth_context_or_response(req)
        self.assertIsNotNone(resp)

        req = RequestFactory().post('/',
                                    data=USERNAME_BODY,
                                    content_type="application/x-www-form-urlencoded",
                                    **USERNAME_META)
        auth_ctx, resp = ACView.get_auth_context_or_response(req)
        self.assertIsNone(resp)
        self.assertEqual(auth_ctx.device_id, '0b1726e825960936d76adf6caf004b97e35a7760')

    def test_get_user_or_response(self):
        req = RequestFactory().post('/',
                                    data=USERNAME_BODY,
                                    content_type="application/x-www-form-urlencoded",
                                    **USERNAME_META)
        auth_ctx, _ = ACView.get_auth_context_or_response(req)

        user, resp = ACView.get_user_or_response(req, auth_ctx)
        self.assertIsNone(user)
        self.assertIsNotNone(resp)

        user = User.objects.create_user('a@a.aa', 'password')
        auth_ctx.user_pk = user.pk
        user, resp = ACView.get_user_or_response(req, auth_ctx)
        self.assertIsNotNone(user)
        self.assertIsNone(resp)


@override_settings(ROOT_URLCONF='acserv.urls',
                   ANYCONNECT=override_auth_header(None))
class IndexACViewTest(TestCase):
    def test_post(self):
        client = Client()

        resp = client.post('/',
                           data=INDEX_BODY,
                           content_type="application/x-www-form-urlencoded",
                           **INDEX_META)
        self.assertTrue('webvpncontext' in resp.cookies)

        auth_ctx = base64_to_proto(resp.cookies['webvpncontext'].value, proto.AuthContext)
        self.assertIsNotNone(auth_ctx.uid, 'webvpncontext是一个pb的message')


@override_settings(ROOT_URLCONF='acserv.urls',
                   ANYCONNECT=override_auth_header(None))
class LoginACViewTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username='a@a.aa',
            email='a@a.aa',
            password='password',
            service_due=timezone.now() + timedelta(100))

    def test_post(self):
        client = Client()

        resp = client.post('/',
                           data=INDEX_BODY,
                           content_type="application/x-www-form-urlencoded",
                           **INDEX_META)

        resp = client.post('/auth/username',
                           data=USERNAME_BODY,
                           content_type="application/x-www-form-urlencoded",
                           )

        resp = client.post('/auth/password',
                           data=PASSWORD_BODY,
                           content_type="application/x-www-form-urlencoded",
                           )
        self.assertContains(resp, "<auth id=\"success\">")

        # 再次登录免输密码
        client2 = Client()
        client2.post('/',
                     data=INDEX_BODY,
                     content_type="application/x-www-form-urlencoded",
                     **INDEX_META)
        resp = client2.post('/auth/username',
                            data=USERNAME_BODY,
                            content_type="application/x-www-form-urlencoded",
                            )
        self.assertContains(resp, "<auth id=\"success\">")

    def test_exceed_max_sessions(self):
        self.user.max_sessions = 1
        self.user.current_sessions = 1
        self.user.save()

        client = Client()
        client.post('/', data=INDEX_BODY, content_type='application/x-www-form-urlencoded', **INDEX_META)
        client.post('/auth/username', data=USERNAME_BODY, content_type='application/x-www-form-urlencoded')
        resp = client.post('/auth/password', data=PASSWORD_BODY, content_type='application/x-www-form-urlencoded')
        self.assertTrue('group_list' in resp.content.decode('utf-8'), '提示要踢一个客户端')

        self.user.max_sessions = 2
        self.user.save()

        client.post('/', data=INDEX_BODY, content_type='application/x-www-form-urlencoded', **INDEX_META)
        client.post('/auth/username', data=USERNAME_BODY, content_type='application/x-www-form-urlencoded')
        resp = client.post('/auth/password', data=PASSWORD_BODY, content_type='application/x-www-form-urlencoded')
        self.assertFalse('group_list' in resp.content.decode('utf-8'), '最大客户端数超过已登录客户端，不提示踢人')


@override_settings(ROOT_URLCONF='acserv.urls',
                   ANYCONNECT=override_auth_header(None))
class KickSessionACViewTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('a@a.aa',
                                             email='a@a.aa',
                                             password='password',
                                             service_due=timezone.now() + timedelta(100))

    def test(self):
        token = 'abc'
        self.user.current_sessions = 2
        self.user.max_sessions = 2
        self.user.token = token
        self.user.save()

        client = Client()
        client.post('/', data=INDEX_BODY, content_type='application/x-www-form-urlencoded', **INDEX_META)
        client.post('/auth/username', data=USERNAME_BODY, content_type='application/x-www-form-urlencoded')
        resp = client.post('/auth/password', data=PASSWORD_BODY, content_type='application/x-www-form-urlencoded')
        self.assertTrue('group_list' in resp.content.decode('utf-8'), '提示要踢一个客户端')

        resp = client.post('/auth/clients',
                           data="group_list={}".format('no such token'),
                           content_type='application/x-www-form-urlencoded')
        self.assertTrue('group_list' in resp.content.decode('utf-8'), '要踢的login_token不对，踢失败，继续显示踢人界面')

        resp = client.post('/auth/clients',
                           data="group_list={}".format(self.user.token),
                           content_type='application/x-www-form-urlencoded')

        self.assertTrue('success' in resp.content.decode('utf-8'), '踢人之后登录成功')
        self.assertEqual(self.user.current_sessions, 2, '去掉一个，加上一个，还是2个')



import base64


def proto_to_base64(m):
    chunk = m.SerializeToString()
    return base64.urlsafe_b64encode(chunk).decode('utf-8')


def base64_to_proto(s, cls):
    chunk = base64.urlsafe_b64decode(s)
    m = cls()
    m.ParseFromString(chunk)
    return m



from django.conf.urls import url

from acserv.views import KickSessionACView, IndexACView, LogoutACView, PasswordACView, UsernameACView

urlpatterns = [
    url('^$', IndexACView.as_view()),
    url('^auth/username$', UsernameACView.as_view()),
    url('^auth/password$', PasswordACView.as_view()),
    url('^auth/sessions', KickSessionACView.as_view()),
    url('^logout$', LogoutACView.as_view()),
]



from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.db.models.signals import post_save
from django.utils import timezone


class User(AbstractUser):
    service_due = models.DateTimeField(null=True)
    max_sessions = models.PositiveIntegerField(default=2)

    def is_in_service(self):
        return self.service_due and self.service_due > timezone.now()


class ClientSession(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="client_sessions")
    device_id = models.CharField(max_length=128, unique=True, null=True)
    app = models.CharField(max_length=32, default='', blank=True)
    os = models.CharField(max_length=32, default='', blank=True)
    ip = models.GenericIPAddressField(default='', blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)

    @classmethod
    def post_save(cls, sender, instance: 'ClientSession', raw, created, **kwargs):
        if raw:
            return


post_save.connect(ClientSession.post_save, sender=ClientSession)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: models.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='models.proto',
  package='models',
  syntax='proto3',
  serialized_pb=_b('\n\x0cmodels.proto\x12\x06models\"V\n\x0eSessionContext\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07user_pk\x18\x02 \x01(\x04\x12\x13\n\x0bservice_due\x18\x03 \x01(\x04\x12\x12\n\ncreated_at\x18\x04 \x01(\x04\"\xd4\x01\n\x0b\x41uthContext\x12\x0b\n\x03uid\x18\x01 \x01(\t\x12\x0f\n\x07user_pk\x18\x02 \x01(\x04\x12\x11\n\tdevice_id\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65vice_type\x18\x04 \x01(\t\x12\x12\n\ndevice_mac\x18\x05 \x01(\t\x12\x13\n\x0b\x64\x65vice_imei\x18\x06 \x01(\t\x12\x10\n\x08platform\x18\x07 \x01(\t\x12\x18\n\x10platform_version\x18\x08 \x01(\t\x12\x16\n\x0e\x63lient_version\x18\t \x01(\t\x12\x12\n\nuser_agent\x18\n \x01(\tb\x06proto3')
)




_SESSIONCONTEXT = _descriptor.Descriptor(
  name='SessionContext',
  full_name='models.SessionContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='models.SessionContext.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='user_pk', full_name='models.SessionContext.user_pk', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='service_due', full_name='models.SessionContext.service_due', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_at', full_name='models.SessionContext.created_at', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24,
  serialized_end=110,
)


_AUTHCONTEXT = _descriptor.Descriptor(
  name='AuthContext',
  full_name='models.AuthContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uid', full_name='models.AuthContext.uid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='user_pk', full_name='models.AuthContext.user_pk', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_id', full_name='models.AuthContext.device_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_type', full_name='models.AuthContext.device_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_mac', full_name='models.AuthContext.device_mac', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_imei', full_name='models.AuthContext.device_imei', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='platform', full_name='models.AuthContext.platform', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='platform_version', full_name='models.AuthContext.platform_version', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_version', full_name='models.AuthContext.client_version', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='user_agent', full_name='models.AuthContext.user_agent', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=113,
  serialized_end=325,
)

DESCRIPTOR.message_types_by_name['SessionContext'] = _SESSIONCONTEXT
DESCRIPTOR.message_types_by_name['AuthContext'] = _AUTHCONTEXT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SessionContext = _reflection.GeneratedProtocolMessageType('SessionContext', (_message.Message,), dict(
  DESCRIPTOR = _SESSIONCONTEXT,
  __module__ = 'models_pb2'
  # @@protoc_insertion_point(class_scope:models.SessionContext)
  ))
_sym_db.RegisterMessage(SessionContext)

AuthContext = _reflection.GeneratedProtocolMessageType('AuthContext', (_message.Message,), dict(
  DESCRIPTOR = _AUTHCONTEXT,
  __module__ = 'models_pb2'
  # @@protoc_insertion_point(class_scope:models.AuthContext)
  ))
_sym_db.RegisterMessage(AuthContext)


# @@protoc_insertion_point(module_scope)



import base64
import json
import uuid
import xml.etree.ElementTree as ET
from datetime import timedelta
from typing import List, Tuple

import pydash as _
from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.functional import SimpleLazyObject
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from acserv import proto
from acserv.helpers import base64_to_proto, proto_to_base64

User = get_user_model()

VERSIONS_CAN_HIDE_USERNAME = [
    "4.2.03013",  # win&mac
    "4.0.05069",  # ios
    "4.0.05057"  # android
]


def get_adapter():
    settings_adapter = _.get(settings, 'ANYCONNECT.ADAPTER')
    if settings_adapter:
        import importlib
        tokens = settings_adapter.split('.')
        pkg = importlib.import_module('.'.join(tokens[:-1]))
        return getattr(pkg, tokens[-1])
    else:
        return BaseAdapter


class BaseAdapter:
    @classmethod
    def can_skip_password(cls, auth_ctx, user) -> bool:
        """ 账号验证后，是否可以跳过密码过程 """
        return False

    @classmethod
    def remove_session(cls, user, device_id) -> str:
        """清理session，返回 id，供go来踢人"""
        pass

    @classmethod
    def make_session_context_and_message(cls, request, auth_ctx, user) -> Tuple[proto.SessionContext, str]:
        """ 身份验证成功后, 判断service_due等, 并生成session_context，acserv会将内容打包放在 webvpn cookie中，供go识别 """
        raise NotImplementedError

    @classmethod
    def get_kick_session_options(cls, user) -> List:
        """ 客户端超过限制后form中的option条目 """
        return [
            # {value, label},
        ]


@method_decorator(csrf_exempt, name='dispatch')
class ACView(View):
    # 为True的view会重新生成auth_context而不从cookie中读取，一般用在Index上
    reset_auth_context = False

    adapter = SimpleLazyObject(get_adapter)  # url引入view时会有顺序问题而报错

    def dispatch(self, request, *args, **kwargs):
        # 如果提供了验证头的配置，就一定要匹配，否则403
        AUTH_HEADER = _.get(settings, 'ANYCONNECT.AUTH_HEADER')
        if AUTH_HEADER:
            for k, v in AUTH_HEADER.items():
                meta_key = 'HTTP_{}'.format(k.replace('-', '_').upper())
                if request.META.get(meta_key) == v:
                    break
            else:
                return HttpResponseForbidden()

        # 在正式处理请求之前，生成或获得auth_ctx
        if self.reset_auth_context:
            auth_ctx = self.make_auth_context(request)
        else:
            auth_ctx, resp = self.get_auth_context_or_response(request)
            if resp:
                return resp

        resp = super(ACView, self).dispatch(request, auth_ctx, *args, **kwargs)

        # 在请求处理之后，追加更新后的auth_ctx
        self.attach_auth_context(resp, auth_ctx)

        return resp

    @classmethod
    def make_auth_context(cls, request):
        auth_context = proto.AuthContext()
        auth_context.uid = str(uuid.uuid4())
        auth_context.user_agent = request.META.get('HTTP_USER_AGENT')

        root = ET.fromstring(request.body)
        node = root.find('device-id')
        if node is not None:
            auth_context.device_id = node.get('unique-id') or ''
            auth_context.device_type = node.get('device-type') or ''
            auth_context.platform = node.text or request.META.get('HTTP_X_acserv_IDENTIFIER_PLATFORM') or ''
            auth_context.platform_version = node.get('platform-version') or ''

        node = root.find('version')
        if node is not None:
            auth_context.client_version = node.text

        node = root.find('phone-id')
        if node is not None:
            auth_context.device_imei = node.text

        node = root.find('mac-address-list')
        if node is not None:
            auth_context.device_mac = ','.join([x.text for x in node.findall('mac-address')])

        return auth_context

    @classmethod
    def get_auth_context_or_response(cls, request):
        try:
            s = request.COOKIES['webvpncontext']
            auth_ctx = base64_to_proto(s, proto.AuthContext)
            assert auth_ctx, 'must valid'

            return auth_ctx, None
        except Exception as e:
            return None, HttpResponse(status=401)

    @classmethod
    def get_user_or_response(cls, request, auth_ctx):
        if not auth_ctx or not auth_ctx.user_pk:
            return None, cls.render_account_form(request,
                                                 message=render_to_string('acserv/messages/account_required.txt'))

        user = User.objects.filter(pk=auth_ctx.user_pk).first()
        if user:
            return user, None
        else:
            return None, cls.render_account_form(request,
                                                 message=render_to_string('acserv/messages/account_not_found.txt'))

    @classmethod
    def handle_auth_succeed(cls, request, auth_ctx, user):
        cls.adapter.remove_session(user, auth_ctx.device_id)

        options = cls.adapter.get_kick_session_options(user)
        if options:
            return cls.render_kick_session_form(
                request, options,
                message=render_to_string('acserv/messages/sessions_too_many.txt'),
                can_hide_username=auth_ctx.client_version in VERSIONS_CAN_HIDE_USERNAME
            )

        session_context, message = cls.adapter.make_session_context_and_message(request, auth_ctx, user)

        if session_context:
            session_key = proto_to_base64(session_context)
        else:
            session_key = 'rejected'  # acserv收到之后会无法解析这个key而拒绝连接

        return cls.render_result_view(request, message, session_key)

    @classmethod
    def render_xml(cls, request, template_name, **context):
        resp = render(request, template_name, context)
        resp["X-transcend-Version"] = '1'
        resp["Content-Type"] = "text/xml"
        return resp

    @classmethod
    def attach_auth_context(cls, response, auth_ctx):
        response.set_cookie('webvpncontext',
                            proto_to_base64(auth_ctx),
                            expires=timezone.now() + timedelta(1),
                            path='/',
                            secure=True)
        return response

    @classmethod
    def render_account_form(cls, request, message):
        return cls.render_xml(request, 'acserv/account.xml',
                              message=message)

    @classmethod
    def render_password_form(cls, request, message):
        return cls.render_xml(request, 'acserv/password.xml',
                              message=message)

    @classmethod
    def render_kick_session_form(cls, request, options, message, can_hide_username=False):
        """ 踢出多余的客户端 """
        return cls.render_xml(request, 'acserv/kick_session.xml',
                              options=options,
                              message=message,
                              can_hide_username=can_hide_username)

    @classmethod
    def render_result_view(cls, request, message, session_key=''):
        resp = cls.render_xml(request, 'acserv/success.xml', message=message)
        resp.set_cookie('webvpn', session_key, secure=True)
        return resp

    @classmethod
    def attach_ac_command(cls, res, kind, payload):
        obj = {"kind": kind, "payload": payload}
        s = base64.urlsafe_b64encode(json.dumps(obj).encode())
        res["X-Ac-Command"] = s
        return s


class IndexACView(ACView):
    reset_auth_context = True

    def post(self, request, auth_ctx):
        if not auth_ctx.device_id:
            return self.render_result_view(request,
                                           message=render_to_string('acserv/messages/index_missing_device_id.txt'))
        return self.render_account_form(request,
                                        message=render_to_string('acserv/messages/account_required.txt'))


class UsernameACView(ACView):
    def post(self, request, auth_ctx):
        # 账号未输入
        username = self.request.POST.get('username', '').lower()
        if not username:
            return self.render_account_form(request,
                                            message=render_to_string('acserv/messages/account_required.txt'))

        # 用户不存在
        user = User.objects.filter(username=username).first()
        if not user:
            return self.render_account_form(request,
                                            message=render_to_string('acserv/messages/account_not_found.txt'))

        auth_ctx.user_pk = user.pk

        # 有上次登录记录的，直接登录
        if self.adapter.can_skip_password(auth_ctx, user):
            return self.handle_auth_succeed(request, auth_ctx, user)

        # 没有登录记录的，提示输入密码
        return self.render_password_form(request,
                                         message=render_to_string('acserv/messages/password_required.txt'))


class PasswordACView(ACView):
    def post(self, request, auth_ctx):
        user, resp = self.get_user_or_response(request, auth_ctx)
        if resp:
            return resp

        # 密码未输
        password = request.POST.get('password')
        if not password:
            return self.render_password_form(request,
                                             message=render_to_string('acserv/messages/password_required.txt'))

        # 密码不正确
        if not user.check_password(password):
            return self.render_password_form(request,
                                             message=render_to_string('acserv/messages/password_wrong.txt'))

        return self.handle_auth_succeed(request, auth_ctx, user)


class KickSessionACView(ACView):
    def post(self, request, auth_ctx):
        user, resp = self.get_user_or_response(request, auth_ctx)
        if resp:
            return resp

        device_id = request.POST.get('group_list', '')
        id = self.adapter.remove_session(user, device_id)  # todo: kick失败的处理逻辑

        res = self.handle_auth_succeed(request, auth_ctx, user)
        self.attach_ac_command(res, "kick_session", id)
        return res


class LogoutACView(ACView):
    # 显示banner时，如果拒绝，就进入这里
    def get(self, request, auth_ctx):
        return HttpResponse('ok')



import timeago
from django.contrib.auth import get_user_model
from django.utils import timezone
from rest_framework import serializers

from sample.models import ClientSession

User = get_user_model()


class ClientSessionSerializer(serializers.ModelSerializer):
    created_time_ago = serializers.SerializerMethodField()

    class Meta:
        model = ClientSession
        fields = ('app', 'os', 'ip', 'created_time_ago', 'device_id')

    def get_os(self, cs: ClientSession):
        return self.get_os_brand(cs.os)

    def get_created_time_ago(self, cs: ClientSession):
        return timeago.format(cs.created_at, timezone.now(), 'zh_CN')



from django.contrib.auth import get_user_model
from django.template.loader import render_to_string
from ipware.ip import get_ip

from acserv import proto
from acserv.views import BaseAdapter
from sample.models import ClientSession
from sample.serializers import ClientSessionSerializer

User = get_user_model()


class ACAdapter(BaseAdapter):
    @classmethod
    def can_skip_password(cls, auth_ctx, user):
        return ClientSession.objects.filter(device_id=auth_ctx.device_id, user=user).exists()

    @classmethod
    def remove_session(cls, user, device_id) -> str:
        cs = user.client_sessions.filter(device_id=device_id).first()
        if cs:
            cs_id = cs.id
            cs.delete()
            return str(cs_id)
        return None

    @classmethod
    def make_session_context_and_message(cls, request, auth_ctx, user):
        if not user.is_in_service():
            return None, '服务过期\n 请到baidu.com充值'

        client_session = user.client_sessions.create(ip=get_ip(request),
                                                     os=auth_ctx.platform + ' ' + auth_ctx.platform_version,
                                                     app='anyconnect')

        if auth_ctx.device_id:
            client_session.device_id = auth_ctx.device_id
            client_session.save()

        session_ctx = proto.SessionContext()
        session_ctx.id = str(client_session.id)
        session_ctx.user_pk = user.pk
        session_ctx.service_due = int(user.service_due.timestamp() * 1000)

        message = render_to_string('acserv/banner.txt',
                                   context={'service_due': user.service_due,
                                            'current_sessions': user.client_sessions.count(),
                                            'max_sessions': user.max_sessions})
        return session_ctx, message

    @classmethod
    def get_kick_session_options(cls, user):
        if user.client_sessions.count() + 1 > user.max_sessions:
            data = ClientSessionSerializer(instance=user.client_sessions, many=True).data
            for client in data:
                client['label'] = "{os} {app} ({created_time_ago})".format(**client)
                client['value'] = client.get('device_id')
            return data
